from unittest import TestCase
from unittest.mock import patch

import requests_mock
from apistar.exceptions import ErrorResponse

from teklia_toolbox.requests import retried_request


@requests_mock.Mocker()
@patch(
    "teklia_toolbox.requests.retried_request.retry.wait",
    0,
)
class TestRetriedRequest(TestCase):
    def test_retried_request_only_500(self, mock):
        mock.get("https://arkindex.teklia.com/api/v1/corpus/notfound/", status_code=404)

        with self.assertRaises(ErrorResponse) as e:
            retried_request("RetrieveCorpus", id="notfound")

        self.assertEqual(e.exception.status_code, 404)
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/notfound/"),
            ],
        )

    def test_retried_request_5_attempts(self, mock):
        mock.get("https://arkindex.teklia.com/api/v1/corpus/corpusid/", status_code=500)

        with self.assertRaises(ErrorResponse) as e:
            retried_request("RetrieveCorpus", id="corpusid")

        self.assertEqual(e.exception.status_code, 500)
        self.assertListEqual(
            [(req.method, req.url) for req in mock.request_history],
            [
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
                ("GET", "https://arkindex.teklia.com/api/v1/corpus/corpusid/"),
            ],
        )



#!/usr/bin/env python
from pathlib import Path

from setuptools import find_packages, setup


def requirements(path):
    assert path.exists(), f"Missing requirements {path}"
    with path.open() as f:
        return list(map(str.strip, f.read().splitlines()))


with Path("VERSION").open() as f:
    VERSION = f.read()

install_requires = requirements(Path("requirements.txt"))

setup(
    name="teklia-toolbox",
    version=VERSION,
    author="Teklia",
    author_email="contact@teklia.com",
    python_requires=">=3.10",
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,
)



import logging
from pathlib import Path
from urllib.parse import urlparse

import requests
import urllib3
from apistar.exceptions import ErrorResponse
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
    wait_fixed,
)

from arkindex import ArkindexClient, options_from_env

logger = logging.getLogger(__name__)


def should_verify_cert(url: str) -> bool:
    """
    Skip SSL certification validation when hitting a development instance
    """
    if not url:
        return True

    host = urlparse(url).netloc
    return not host.endswith("ark.localhost")


def _get_arkindex_client() -> ArkindexClient:
    options = options_from_env()

    # Skip SSL verification in Arkindex API client for local development hosts
    verify = should_verify_cert(options.get("base_url"))
    if not verify:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        logger.warn("SSL certificate verification is disabled for Arkindex API calls")

    return ArkindexClient(verify=verify, **options)


DEFAULT_CLIENT = _get_arkindex_client()


def _is_500_error(exc: Exception) -> bool:
    """
    Check if an Arkindex API error is a 50x
    This is used to retry most API calls implemented here
    """
    if not isinstance(exc, ErrorResponse):
        return False

    return 500 <= exc.status_code < 600


@retry(
    retry=retry_if_exception(_is_500_error),
    wait=wait_exponential(multiplier=2, min=3),
    reraise=True,
    stop=stop_after_attempt(5),
    before_sleep=before_sleep_log(logger, logging.INFO),
)
def retried_request(*args, **kwargs):
    """
    Proxy all Arkindex API requests with a retry mechanism
    in case of 50X errors
    The same API call will be retried 5 times, with an exponential sleep time
    going through 3, 4, 8 and 16 seconds of wait between call.
    If the 5th call still gives a 50x, the exception is re-raised
    and the caller should catch it
    Log messages are displayed before sleeping (when at least one exception occurred)
    """
    return DEFAULT_CLIENT.request(*args, **kwargs)


# Time to wait before retrying the IIIF image information fetching
HTTP_GET_RETRY_BACKOFF = 10

DOWNLOAD_CHUNK_SIZE = 8192


@retry(
    reraise=True,
    retry=retry_if_exception_type(requests.RequestException),
    stop=stop_after_attempt(3),
    wait=wait_fixed(HTTP_GET_RETRY_BACKOFF),
)
def download_file(url: str, path: Path) -> None:
    """
    Download a URL into a local path, retrying if necessary
    """
    with requests.get(url, stream=True, verify=should_verify_cert(url)) as r:
        r.raise_for_status()
        with path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                if chunk:  # Ignore empty chunks
                    f.write(chunk)
